/******************************************************************************
 *                castor/db/newora/OraCommonSvc.hpp
 *
 * This file is part of the Castor project.
 * See http://castor.web.cern.ch/castor
 *
 * Copyright (C) 2003  CERN
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * @(#)$RCSfile: OraCommonSvc.hpp,v $ $Revision: 1.14 $ $Release$ $Date: 2009/03/26 10:59:47 $ $Author: itglp $
 *
 * Implementation of the ICommonSvc for Oracle/CDBC
 *
 * @author Sebastien Ponce
 *****************************************************************************/

#ifndef ORA_ORACOMMONSVC_HPP
#define ORA_ORACOMMONSVC_HPP 1

// Include Files
#include "castor/BaseSvc.hpp"
#include "castor/db/DbBaseObj.hpp"
#include "castor/db/newora/OraCnvSvc.hpp"
#include "castor/stager/ICommonSvc.hpp"
#include "occi.h"
#include <vector>

namespace castor {

  namespace db {

    namespace ora {

      /**
       * Implementation of the ICommonSvc for Oracle
       */
      class OraCommonSvc : public BaseSvc,
                           public DbBaseObj,
                           public virtual castor::stager::ICommonSvc {

      public:

        /**
         * default constructor
         */
        OraCommonSvc(const std::string name);

        /**
         * default destructor
         */
        virtual ~OraCommonSvc() throw();

        /**
         * Get the service id
         */
        virtual const unsigned int id() const;

        /**
         * Get the service id
         */
        static const unsigned int ID();

        /**
         * Reset the converter statements.
         */
        void reset() throw ();

      public:

        /**
         * Selects the next request a stager service should deal with.
         * Selects a Request in START status and deletes it from the
         * NewRequests helper table to avoid double processing.
         * @param service the stager service that will process the Request
         * @return the Request to process
         * @exception Exception in case of error
         */
        virtual castor::stager::Request* requestToDo(std::string service)
          throw (castor::exception::Exception);

        /**
         * Retrieves a tape from the database based on its vid,
         * side and tpmode. If no tape is found, creates one
         * This method holds a lock on the row only on creation
         * of a new tape. It is the* responsability of the caller
         * to commit the transaction.
         * @param vid the vid of the tape
         * @param side the side of the tape
         * @param tpmode the tpmode of the tape
         * @return the tape. the return value can never be 0
         * @exception Exception in case of error (no tape found,
         * several tapes found, DB problem, etc...)
         */
        castor::stager::Tape* selectTape(const std::string vid,
                                         const int side,
                                         const int tpmode)
          throw (castor::exception::Exception);

        /**
         * Retrieves a SvcClass from the database based on its name.
         * Caller is in charge of the deletion of the allocated object
         * @param name the name of the SvcClass
         * @return the SvcClass, or 0 if none found
         * @exception Exception in case of error
         */
        virtual castor::stager::SvcClass* selectSvcClass
        (const std::string name)
          throw (castor::exception::Exception);

        /**
         * Retrieves a FileClass from the database based on its name.
         * Caller is in charge of the deletion of the allocated object
         * @param name the name of the FileClass
         * @return the FileClass, or 0 if none found
         * @exception Exception in case of error
         */
        virtual castor::stager::FileClass* selectFileClass
        (const std::string name)
          throw (castor::exception::Exception);

      protected:

        /**
         * helper method to commit
         */
        virtual void commit();

        /**
         * helper method to rollback
         */
        virtual void rollback();

        /**
         * Helper method to handle exceptions - see OraCnvSvc
         * @param e an Oracle exception
         */
        void handleException(oracle::occi::SQLException& e);

        /**
         * helper method to create Oracle statement
         */
        virtual oracle::occi::Statement* createStatement(const std::string& stmtString)
          throw (castor::exception::Exception);

        /**
         * helper method to delete Oracle statement
         */
        virtual void deleteStatement(oracle::occi::Statement* stmt)
          throw (castor::exception::Exception);

      private:

        /// SQL statement for function requestToDo
        static const std::string s_requestToDoStatementString;

        /// SQL statement object for function requestToDo
        oracle::occi::Statement *m_requestToDoStatement;

        /// SQL statement for function selectTape
        static const std::string s_selectTapeStatementString;

        /// SQL statement object for function selectTape
        oracle::occi::Statement *m_selectTapeStatement;

        /// SQL statement for function selectSvcClass
        static const std::string s_selectSvcClassStatementString;

        /// SQL statement object for function selectSvcClass
        oracle::occi::Statement *m_selectSvcClassStatement;

        /// SQL statement for function selectFileClass
        static const std::string s_selectFileClassStatementString;

        /// SQL statement object for function selectFileClass
        oracle::occi::Statement *m_selectFileClassStatement;

      }; // end of class OraCommonSvc

    } // end of namespace ora

  } // end of namespace db

} // end of namespace castor

#endif // ORA_ORACOMMONSVC_HPP
