/******************************************************************************
 *              dlf_2.1.9-5_to_2.1.9-6.sql
 *
 * This file is part of the Castor project.
 * See http://castor.web.cern.ch/castor
 *
 * Copyright (C) 2003  CERN
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * This script upgrades a CASTOR v2.1.9-5 DLF database to v2.1.9-6
 *
 * @author Castor Dev team, castor-dev@cern.ch
 *****************************************************************************/

/* Stop on errors */
WHENEVER SQLERROR EXIT FAILURE
DECLARE
  unused VARCHAR2(30);
BEGIN
  UPDATE UpgradeLog
     SET failureCount = failureCount + 1
   WHERE schemaVersion = '2_1_9_3'
     AND release = '2_1_9_6'
     AND state != 'COMPLETE';
  COMMIT;
EXCEPTION WHEN NO_DATA_FOUND THEN
  NULL;
END;
/

/* Version cross check and update */
DECLARE
  unused VARCHAR(100);
BEGIN
  SELECT release INTO unused FROM CastorVersion
   WHERE release LIKE '2_1_9_5%';
EXCEPTION WHEN NO_DATA_FOUND THEN
  -- Error, we can't apply this script
  raise_application_error(-20000, 'PL/SQL release mismatch. Please run previous upgrade scripts before this one.');
END;
/

INSERT INTO UpgradeLog (schemaVersion, release, type)
VALUES ('2_1_9_3', '2_1_9_6', 'TRANSPARENT');
COMMIT;

/* Schema changes go here */
/**************************/

/* New constraints */
ALTER TABLE dlf_config MODIFY
  (value CONSTRAINT NN_Config_Value NOT NULL, description CONSTRAINT NN_Config_Description NOT NULL);

ALTER TABLE dlf_severities MODIFY
  (sev_no CONSTRAINT PK_Severities_Sev_No PRIMARY KEY CONSTRAINT NN_Severities_Sev_No NOT NULL,
   sev_name CONSTRAINT NN_Severities_Sev_Name NOT NULL);
ALTER TABLE dlf_severities DROP CONSTRAINT UN_Severities_Sev_NoName;
DROP INDEX UN_Severities_Sev_NoName;

ALTER TABLE dlf_facilities DROP CONSTRAINT UN_Facilities_Fac_Name;
ALTER TABLE dlf_facilities DROP CONSTRAINT UN_Facilities_Fac_No;
ALTER TABLE dlf_facilities MODIFY
  (fac_no CONSTRAINT PK_Facilities_Fac_No PRIMARY KEY CONSTRAINT NN_Facilities_Fac_No NOT NULL,
   fac_name CONSTRAINT NN_Facilities_Fac_Name NOT NULL);
ALTER INDEX UN_Facilities_Fac_No RENAME TO PK_Facilities_Fac_No;
DROP INDEX UN_Facilities_Fac_Name;

ALTER TABLE dlf_msg_texts MODIFY
  (fac_no CONSTRAINT NN_Msg_Texts_Fac_No NOT NULL, msg_no CONSTRAINT NN_Msg_Texts_Msg_No NOT NULL,
   msg_text CONSTRAINT NN_Msg_Texts_Msg_Text NOT NULL);

ALTER TABLE dlf_msg_texts
  ADD CONSTRAINT PK_Msg_Texts_FacMsgNo PRIMARY KEY (fac_no, msg_no);
ALTER INDEX UN_Msg_Texts_FacMsgNo RENAME TO PK_Msg_Texts_FacMsgNo;

ALTER TABLE dlf_host_map DROP CONSTRAINT UN_Host_Map_HostID; 
ALTER TABLE dlf_host_map DROP CONSTRAINT UN_Host_Map_Hostname;
ALTER TABLE dlf_host_map MODIFY
  (hostid CONSTRAINT PK_Host_Map_HostID PRIMARY KEY CONSTRAINT NN_Host_Map_HostID NOT NULL,
   hostname CONSTRAINT NN_Host_Map_HostName NOT NULL);
ALTER INDEX UN_Host_Map_HostID RENAME TO PK_Host_Map_HostID;
DROP INDEX UN_Host_Map_HostName;

ALTER TABLE dlf_nshost_map DROP CONSTRAINT UN_NSHost_Map_NSHostID;
ALTER TABLE dlf_nshost_map DROP CONSTRAINT UN_NSHost_Map_NSHostname;
ALTER TABLE dlf_nshost_map MODIFY
  (nshostid CONSTRAINT PK_NSHost_Map_NSHostID PRIMARY KEY CONSTRAINT NN_NSHost_Map_NSHostID NOT NULL,
   nshostname CONSTRAINT NN_NSHost_Map_NSHostName NOT NULL);
ALTER INDEX UN_NSHost_Map_NSHostID RENAME TO PK_NSHost_Map_NSHostID;
DROP INDEX UN_NSHost_Map_NsHostName;

/* Disable all database jobs */
BEGIN
  FOR a IN (SELECT job_name FROM user_scheduler_jobs)
  LOOP
    dbms_scheduler.disable(a.job_name);
  END LOOP;
END;
/

/* Flag all tablespaces as online */
DECLARE
  username VARCHAR2(2048) := SYS_CONTEXT('USERENV', 'CURRENT_USER');
BEGIN
  FOR a IN (SELECT * FROM user_tablespaces
             WHERE tablespace_name LIKE CONCAT('DLF_%_', username)
               AND status = 'READ ONLY')
  LOOP
    EXECUTE IMMEDIATE 'ALTER TABLESPACE '||a.tablespace_name||' READ WRITE';
  END LOOP;
END;
/

/* Recreate the I_Messages_Facility index as a BITMAP one */
DROP INDEX I_Messages_Facility;
CREATE BITMAP INDEX I_Messages_Facility ON dlf_messages (facility) LOCAL UNUSABLE NOLOGGING;

/* Add BITMAP indexes on severity and msgno to improve query performance */
CREATE BITMAP INDEX I_Messages_Severity ON dlf_messages (severity) LOCAL UNUSABLE NOLOGGING;
CREATE BITMAP INDEX I_Messages_MsgNo ON dlf_messages (msg_no) LOCAL UNUSABLE NOLOGGING;

/* Rebuild all UNUSABLE indexes */
BEGIN
  FOR a IN (SELECT index_name, partition_name, tablespace_name
              FROM user_ind_partitions
             WHERE index_name IN
              ('I_MESSAGES_FACILITY', 'I_MESSAGES_SEVERITY', 'I_MESSAGES_MSGNO')
               AND status = 'UNUSABLE'
             ORDER BY partition_name DESC, index_name ASC)
  LOOP
    EXECUTE IMMEDIATE 'ALTER INDEX '||a.index_name||' REBUILD PARTITION '||
                       a.partition_name||' TABLESPACE '||a.tablespace_name;
  END LOOP;
END;
/

/* Update and revalidation of PL-SQL code */
/******************************************/

/* PL/SQL method implementing createPartition */
CREATE OR REPLACE PROCEDURE createPartitions
AS
  username VARCHAR2(2048);
  partitionMax NUMBER;
  tableSpaceName VARCHAR2(2048);
  highValue DATE;
  cnt NUMBER;
BEGIN
  -- Set the nls_date_format
  EXECUTE IMMEDIATE 'ALTER SESSION SET NLS_DATE_FORMAT = "DD-MON-YYYY"';

  -- Extract the name of the current user running the PL/SQL procedure. This name
  -- will be used within the tablespace names.
  SELECT SYS_CONTEXT('USERENV', 'CURRENT_USER')
    INTO username
    FROM dual;

  -- Loop over all partitioned tables
  FOR a IN (SELECT DISTINCT(table_name)
              FROM user_tab_partitions
             ORDER BY table_name)
  LOOP
    -- Determine the high value on which to split the MAX_VALUE partition of all
    -- tables
    SELECT max(substr(partition_name, 3, 10))
      INTO partitionMax
      FROM user_tab_partitions
     WHERE partition_name <> 'MAX_VALUE'
       AND table_name = a.table_name;

    partitionMax := TO_NUMBER(
      TO_CHAR(TRUNC(TO_DATE(partitionMax, 'YYYYMMDD') + 1), 'YYYYMMDD'));

    -- If this is the first execution there will be no high value, so set it to
    -- today
    IF partitionMax IS NULL THEN
      partitionMax := TO_NUMBER(TO_CHAR(SYSDATE, 'YYYYMMDD'));
    END IF;

    -- Create partition
    FOR b IN (SELECT TO_DATE(partitionMax, 'YYYYMMDD') + rownum - 1 value
                FROM all_objects
               WHERE rownum <=
                     TO_DATE(TO_CHAR(SYSDATE + 7, 'YYYYMMDD'), 'YYYYMMDD') -
                     TO_DATE(partitionMax, 'YYYYMMDD') + 1)
    LOOP
      -- To improve data management each daily partition has its own tablespace
      -- http://www.oracle.com/technology/oramag/oracle/06-sep/o56partition.html

      -- Check if a new tablespace is required before creating the partition
      tableSpaceName := 'DLF_'||TO_CHAR(b.value, 'YYYYMMDD')||'_'||username;
      SELECT count(*) INTO cnt
        FROM user_tablespaces
       WHERE tablespace_name = tableSpaceName;

      IF cnt = 0 THEN
        EXECUTE IMMEDIATE 'CREATE BIGFILE TABLESPACE '||tableSpaceName||'
                           DATAFILE SIZE 100M
                           AUTOEXTEND ON NEXT 200M';
      END IF;

      -- If the tablespace is read only, alter its status to read write for this
      -- operation.
      FOR d IN (SELECT tablespace_name FROM user_tablespaces
                 WHERE tablespace_name = tableSpaceName
                   AND status = 'READ ONLY')
      LOOP
        EXECUTE IMMEDIATE 'ALTER TABLESPACE '||d.tablespace_name||' READ WRITE';
      END LOOP;

      highValue := TRUNC(b.value + 1);
      EXECUTE IMMEDIATE 'ALTER TABLE '||a.table_name||'
                         SPLIT PARTITION MAX_VALUE
                         AT    ('''||TO_CHAR(highValue, 'DD-MON-YYYY')||''')
                         INTO  (PARTITION P_'||TO_CHAR(b.value, 'YYYYMMDD')||'
                                TABLESPACE '||tableSpaceName||',
                                PARTITION MAX_VALUE)
                         UPDATE INDEXES';

      -- Move indexes to the correct tablespace
      FOR c IN (SELECT index_name
                  FROM user_indexes
                 WHERE table_name = a.table_name)
      LOOP
        EXECUTE IMMEDIATE 'ALTER INDEX '||c.index_name||'
                           REBUILD PARTITION P_'||TO_CHAR(b.value, 'YYYYMMDD')||'
                           TABLESPACE '||tableSpaceName;
      END LOOP;
    END LOOP;
  END LOOP;
END;
/


/* PL/SQL method implementing dropPartitions */
CREATE OR REPLACE PROCEDURE dropPartitions (expiry IN NUMBER)
AS
  username VARCHAR2(2048);
  expiryTime NUMBER;
BEGIN
  -- Set the nls_date_format
  EXECUTE IMMEDIATE 'ALTER SESSION SET NLS_DATE_FORMAT = "DD-MON-YYYY"';

  -- Extract the name of the current user running the PL/SQL procedure. This name
  -- will be used within the tablespace names.
  SELECT SYS_CONTEXT('USERENV', 'CURRENT_USER')
    INTO username
    FROM dual;

  -- Extract configurable expiry time
  expiryTime := expiry;
  IF expiryTime = -1 THEN
    SELECT TO_NUMBER(value) INTO expiryTime
      FROM dlf_config
     WHERE name = 'expiry';
  END IF;

  -- Drop partitions across all tables
  FOR a IN (SELECT *
              FROM user_tab_partitions
             WHERE partition_name < concat('P_', TO_CHAR(SYSDATE - expiryTime, 'YYYYMMDD'))
               AND partition_name <> 'MAX_VALUE'
             ORDER BY partition_name DESC)
  LOOP
    EXECUTE IMMEDIATE 'ALTER TABLE '||a.table_name||'
                       DROP PARTITION '||a.partition_name;
  END LOOP;

  -- Drop tablespaces
  FOR a IN (SELECT tablespace_name
              FROM user_tablespaces
             WHERE status <> 'OFFLINE'
               AND tablespace_name LIKE 'DLF_%_'||username
               AND tablespace_name < 'DLF_'||TO_CHAR(SYSDATE - expiryTime, 'YYYYMMDD')||'_'||username
             ORDER BY tablespace_name ASC)
  LOOP
    EXECUTE IMMEDIATE 'ALTER TABLESPACE '||a.tablespace_name||' OFFLINE';
    EXECUTE IMMEDIATE 'DROP TABLESPACE '||a.tablespace_name||'
                       INCLUDING CONTENTS AND DATAFILES';
  END LOOP;

  -- Set the status of tablespaces older then 2 days to read only.
  FOR a IN (SELECT tablespace_name
              FROM user_tablespaces
             WHERE status <> 'OFFLINE'
	       AND status <> 'READ ONLY'
               AND tablespace_name LIKE 'DLF_%_'||username
               AND tablespace_name < 'DLF_'||TO_CHAR(SYSDATE - 2, 'YYYYMMDD')||'_'||username
             ORDER BY tablespace_name ASC)
  LOOP
    EXECUTE IMMEDIATE 'ALTER TABLESPACE '||a.tablespace_name||' READ ONLY';
  END LOOP;
END;
/


/* Scheduler jobs */
BEGIN
  -- Remove scheduler jobs before recreation
  FOR a IN (SELECT job_name FROM user_scheduler_jobs)
  LOOP
    DBMS_SCHEDULER.DROP_JOB(a.job_name, TRUE);
  END LOOP;

  -- Create a db job to be run every day and create new partitions
  DBMS_SCHEDULER.CREATE_JOB(
      JOB_NAME        => 'createPartitionsJob',
      JOB_TYPE        => 'STORED_PROCEDURE',
      JOB_ACTION      => 'createPartitions',
      JOB_CLASS       => 'DLF_JOB_CLASS',
      START_DATE      => TRUNC(SYSDATE) + 1/24,
      REPEAT_INTERVAL => 'FREQ=DAILY',
      ENABLED         => TRUE,
      COMMENTS        => 'Daily partitioning creation');

  -- Create a db job to be run every day and drop old data from the database
  DBMS_SCHEDULER.CREATE_JOB(
      JOB_NAME        => 'dropPartitionsJob',
      JOB_TYPE        => 'PLSQL_BLOCK',
      JOB_ACTION      => 'BEGIN dropPartitions(-1); END;',
      JOB_CLASS       => 'DLF_JOB_CLASS',
      START_DATE      => TRUNC(SYSDATE) + 2/24,
      REPEAT_INTERVAL => 'FREQ=DAILY',
      ENABLED         => TRUE,
      COMMENTS        => 'Daily data removal');
END;
/

/* dropPartitions replaces the archiveData procedure */
DROP PROCEDURE archiveData;

/* Recompile all invalid procedures, triggers and functions */
/************************************************************/
BEGIN
  FOR a IN (SELECT object_name, object_type
              FROM user_objects
             WHERE object_type IN ('PROCEDURE', 'TRIGGER', 'FUNCTION')
               AND status = 'INVALID')
  LOOP
    IF a.object_type = 'PROCEDURE' THEN
      EXECUTE IMMEDIATE 'ALTER PROCEDURE '||a.object_name||' COMPILE';
    ELSIF a.object_type = 'TRIGGER' THEN
      EXECUTE IMMEDIATE 'ALTER TRIGGER '||a.object_name||' COMPILE';
    ELSE
      EXECUTE IMMEDIATE 'ALTER FUNCTION '||a.object_name||' COMPILE';
    END IF;
  END LOOP;
END;
/

/* Flag the schema upgrade as COMPLETE */
/***************************************/
UPDATE UpgradeLog SET endDate = sysdate, state = 'COMPLETE'
 WHERE release = '2_1_9_6';
COMMIT;
