/******************************************************************************
 *              mon_2.1.9-6_to_2.1.9-7.sql
 *
 * This file is part of the Castor project.
 * See http://castor.web.cern.ch/castor
 *
 * Copyright (C) 2003  CERN
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * This script upgrades a CASTOR v2.1.9-6 MONITORING database to v2.1.9-7
 *
 * @author Castor Dev team, castor-dev@cern.ch
 *****************************************************************************/

/* Stop on errors */
WHENEVER SQLERROR EXIT FAILURE
DECLARE
  unused VARCHAR2(30);
BEGIN
  UPDATE UpgradeLog
     SET failureCount = failureCount + 1
   WHERE schemaVersion = '2_1_9_7'
     AND release = '2_1_9_7'
     AND state != 'COMPLETE';
  COMMIT;
EXCEPTION WHEN NO_DATA_FOUND THEN
  NULL;
END;
/

/* Version cross check and update */
DECLARE
  unused VARCHAR(100);
BEGIN
  SELECT release INTO unused FROM CastorVersion
   WHERE release LIKE '2_1_9_6%';
EXCEPTION WHEN NO_DATA_FOUND THEN
  -- Error, we can't apply this script
  raise_application_error(-20000, 'PL/SQL release mismatch. Please run previous upgrade scripts before this one.');
END;
/

/* Determine the DLF schema that the monitoring procedures should run against */
UNDEF dlfschema
ACCEPT dlfschema DEFAULT castor_dlf PROMPT 'Enter the DLF schema to run monitoring queries against: (castor_dlf) ';
SET VER OFF

/* Check that the executing accounting can see the DLF base tables */
DECLARE
  unused VARCHAR2(2048);
BEGIN
  -- Check that the user exists
  BEGIN
    SELECT username INTO unused
      FROM all_users
     WHERE username = upper('&&dlfschema');
  EXCEPTION WHEN NO_DATA_FOUND THEN
    raise_application_error(-20000, 'User &dlfschema does not exist');
  END;
  -- Check that the correct grants are present
  BEGIN
    SELECT owner INTO unused
      FROM all_views
     WHERE owner = upper('&&dlfschema')
       AND view_name = 'CASTORVERSION';
  EXCEPTION WHEN NO_DATA_FOUND THEN
    raise_application_error(-20001, 'Unable to access the &dlfschema..CastorVersion. Check that the correct grants have been issued!');
  END;
END;
/

INSERT INTO UpgradeLog (schemaVersion, release, type)
VALUES ('2_1_9_7', '2_1_9_7', 'TRANSPARENT');
COMMIT;

/* Schema changes go here */
/**************************/

/* SQL statement for table LatencyStats */
DROP TABLE LatencyStats;
CREATE TABLE LatencyStats (timestamp DATE CONSTRAINT NN_LatencyStats_ts NOT NULL, interval NUMBER, requestType VARCHAR2(255), svcClass VARCHAR2(255), protocol VARCHAR2(255), started NUMBER(*,3), minLatencyTime NUMBER(*,3), maxLatencyTime NUMBER(*,3), avgLatencyTime NUMBER(*,3), stddevLatencyTime NUMBER(*,3), medianLatencyTime NUMBER(*,3))
  PARTITION BY RANGE (timestamp) (PARTITION MAX_VALUE VALUES LESS THAN (MAXVALUE));

/* Update and revalidation of PL-SQL code */
/******************************************/

/* PL/SQL method implementing statsLatency
 *
 * Provides statistics on the amount of time a user has had to wait since their
 * request was entered into the system and it actually being served. The
 * returned data is broken down by request type, service class and protocol.
 */
CREATE OR REPLACE PROCEDURE statsLatency (now IN DATE, interval IN NUMBER) AS
BEGIN
  -- Stats table: LatencyStats
  -- Frequency: 5 minutes
  INSERT INTO LatencyStats
    (timestamp, interval, requestType, svcClass, protocol, started,
     minLatencyTime, maxLatencyTime, avgLatencyTime, stddevLatencyTime,
     medianLatencyTime)
    -- Gather data
    SELECT now - 5/1440 timestamp, interval, nvl(type, '-') requestType,
           svcClass, protocol, (count(*) / interval) started,
           min(waitTime)         minLatencyTime,
           max(waitTime)         maxLatencyTime,
           avg(waitTime)         avgLatencyTime,
           stddev_pop(waitTime)  stddevLatencyTime,
           median(waitTime)      medianLatencyTime
      FROM (
        SELECT waitTime,
               max(decode(params.name, 'Type',     params.value, NULL)) type,
               max(decode(params.name, 'Protocol', params.value, NULL)) protocol,
               max(decode(params.name, 'SvcClass', params.value, NULL)) svcClass
          FROM (
            -- Extract the totalWaitTime for all stagerjob's or d2dtransfer's
            -- which have started.
            SELECT params.id, params.value waitTime
              FROM &dlfschema..dlf_messages messages,
                   &dlfschema..dlf_num_param_values params
             WHERE messages.id = params.id
               AND messages.severity = 8     -- Info
               AND ((messages.facility = 26  -- stagerjob
               AND   messages.msg_no = 20)   -- Job started
                OR  (messages.facility = 23  -- d2dtransfer
               AND   messages.msg_no = 25))  -- DiskCopyTransfer started
               AND messages.timestamp >  now - 10/1440
               AND messages.timestamp <= now - 5/1440
               AND params.name = 'TotalWaitTime'
               AND params.timestamp >  now - 10/1440
               AND params.timestamp <= now - 5/1440
           ) results
         -- Attach the type, protocol and service class values
         INNER JOIN &dlfschema..dlf_str_param_values params
            ON results.id = params.id
           AND params.name IN ('Type', 'Protocol', 'SvcClass')
           AND params.timestamp >  now - 10/1440
           AND params.timestamp <= now - 5/1440
         GROUP BY waitTime)
     WHERE svcClass IS NOT NULL
   GROUP BY GROUPING SETS (svcclass, type), (svcclass, protocol);
END;
/

/* Recompile all invalid procedures, triggers and functions */
/************************************************************/
BEGIN
  FOR a IN (SELECT object_name, object_type
              FROM user_objects
             WHERE object_type IN ('PROCEDURE', 'TRIGGER', 'FUNCTION')
               AND status = 'INVALID')
  LOOP
    IF a.object_type = 'PROCEDURE' THEN
      EXECUTE IMMEDIATE 'ALTER PROCEDURE '||a.object_name||' COMPILE';
    ELSIF a.object_type = 'TRIGGER' THEN
      EXECUTE IMMEDIATE 'ALTER TRIGGER '||a.object_name||' COMPILE';
    ELSE
      EXECUTE IMMEDIATE 'ALTER FUNCTION '||a.object_name||' COMPILE';
    END IF;
  END LOOP;
END;
/

/* Flag the schema upgrade as COMPLETE */
/***************************************/
UPDATE UpgradeLog SET endDate = sysdate, state = 'COMPLETE'
 WHERE release = '2_1_9_7';
COMMIT;
